/* This file is part of mailfrom filter.
   Copyright (C) 2005, Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
   MA 02110-1301 USA */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <mailutils/mailutils.h>
#include "mailfrom.h"

#ifdef USE_DBM

struct cache_result {
	time_t timestamp;
	mf_status status;
};

static char *dbname = DEFAULT_DATABASE;

#define EXPIRE_INTERVAL(s) \
 ((s) == mf_success ? positive_expire_interval : negative_expire_interval)

mf_status
cache_get(char *email)
{
	mf_status rc;
	DBM_FILE db;
	DBM_DATUM key;
	DBM_DATUM contents;

	debug1(50, "getting cache info for %s", email);
	if (mu_dbm_open(dbname, &db, MU_STREAM_RDWR, 0600)) {
		mu_error("mu_dbm_open(%s) failed: %s", dbname,
			 mu_strerror(errno));
		return mf_failure;
	}
	memset (&key, 0, sizeof key);
	memset (&contents, 0, sizeof contents);
	MU_DATUM_PTR (key) = email;
	MU_DATUM_SIZE (key) = strlen (email);
	if (mu_dbm_fetch(db, key, &contents) == 0) {
		struct cache_result *res = (struct cache_result *)
			                     MU_DATUM_PTR(contents);
		time_t t = time(NULL);
		debug2(50, "found status: %d, time: %lu",
			  res->status, res->timestamp);
		
		if (t - res->timestamp > EXPIRE_INTERVAL(res->status)) {
			debug1(50, "removing expired entry for %s",email);
			if (mu_dbm_delete (db, key))
				mu_error ("Cannot remove record for %s",
					  email);
			rc = mf_failure;
		} else {
			rc = res->status;
		}
	} else 
		rc = mf_failure;
	mu_dbm_datum_free(&contents);
	mu_dbm_close(db);
	return rc;
}

void
cache_insert(char *email, mf_status rc)
{
	DBM_FILE db;
	DBM_DATUM key;
	DBM_DATUM contents;
	struct cache_result res;

	debug3(50,"inserting cache info for %s. status=%d, time=%lu",
		  email, rc, time(NULL));
	if (mu_dbm_open(dbname, &db, MU_STREAM_RDWR, 0600)) {
		mu_error("mu_dbm_open(%s) failed: %s", dbname,
			 mu_strerror(errno));
		return;
	}
	memset(&key, 0, sizeof key);
	memset(&contents, 0, sizeof contents);
	MU_DATUM_PTR(key) = email;
	MU_DATUM_SIZE(key) = strlen (email);
	time(&res.timestamp);
	res.status = rc;
	MU_DATUM_PTR(contents) = (void*)&res;
	MU_DATUM_SIZE(contents) = sizeof(res);
	if (mu_dbm_insert(db, key, contents, 1))
		mu_error ("Cannot insert datum");
	mu_dbm_close(db);
}

void
cache_delete(char *email)
{
	debug1(50,"deleting cache info for %s", email);
	db_delete(dbname, email);
}

static const char *
format_status(mf_status status)
{
	switch (status) {
	case mf_success:
		return "success";
	case mf_not_found:
		return "not_found";
	case mf_failure:
		return "failure";
	case mf_temp_failure:
		return "temp_failure";
	}
	return "unknown";
}

static void
print_cache_item(const char *email, size_t size, const void *content)
{
	const struct cache_result *res = content;
	
	printf("%*.*s", size, size, email);
	if (size < 24) {
		size = 24 - size;
		printf("%*.*s", size, size, "");
	}
	printf(" %10.10s %s",
	       format_status(res->status), ctime(&res->timestamp));
}

void
cache_list_item(char *email)
{
	db_list_item(dbname, email, print_cache_item);
}

void
cache_list_db()
{
	db_list(dbname, print_cache_item);
}

static int
cache_expire_item(const void *content)
{
	const struct cache_result *res = (struct cache_result *)content;
	return !res || time(NULL) - res->timestamp > EXPIRE_INTERVAL(res->status);
}

void
cache_expire_db()
{
	db_expire(dbname, cache_expire_item);
}

mf_status
cache_get2(char *email, char *client_addr)
{
	mf_status rc = mf_failure;
	size_t size = strlen(email) + 1 + strlen(client_addr) + 1;
	char *key = malloc(size);
	if (key) {
		strcat(strcat(strcpy(key, email), ":"), client_addr);
		rc = cache_get(key);
		free(key);
	}
	return rc;
}

void
cache_insert2(char *email, char *client_addr, mf_status rc)
{
	size_t size = strlen(email) + 1 + strlen(client_addr) + 1;
	char *key = malloc(size);
	if (key) {
		strcat(strcat(strcpy(key, email), ":"), client_addr);
		cache_insert(key, rc);
		free(key);
	}
}

#endif
